from anvil.errors import EmptyRegionFile, InvalidFileType
from anvil.empty_region import EmptyRegion
from anvil.empty_chunk import EmptyChunk
import context as _
import pytest
from anvil import Region
import io
import secrets

# TODO: Implement tests for anvil/region.py
#
# TestRegion class:
#   - A test for chunk_location() to ensure it returns the correct offset and sector count for a given chunk.
#   - A test for chunk_location() on a non-existent chunk to ensure it returns `(0, 0)`.
#   - A test for chunk_data() on a non-existent chunk to ensure it returns `None`.

def test_from_filename(tmp_path) -> None:
    filename = tmp_path / "region.mca"
    contents = secrets.token_bytes()

    with open(filename, 'wb') as f:
        f.write(contents)

    region = Region.from_file(str(filename))
    assert region.data == contents

def test_from_filelike() -> None:
    contents = secrets.token_bytes()
    filelike = io.BytesIO(contents)

    region = Region.from_file(filelike)
    assert region.data == contents

@pytest.mark.parametrize("_,value", [
    ('int', 123),
    ('list', ['not', 'a', 'file']),
    ('None_type', None),
    ('dict', {'key': 'value'}),
    ('float', 420.69),
    ('object', object()),
    ('set', {'some', 'set'})
])

def test_from_file_raises_InvalidFileType(_: str, value: object) -> None:
    with pytest.raises(InvalidFileType):
        Region.from_file(value) # type: ignore

def test_chunk_location_existing_chunk() -> None:
    # Create region with a chunk
    empty_region = EmptyRegion(0, 0)
    chunk = EmptyChunk(0, 0)
    empty_region.add_chunk(chunk)

    # Convert to Region
    region_bytes = empty_region.save()
    region = Region(region_bytes)

    result = region.chunk_location(0, 0)
    if result is None:
        offset, sectors = 0, 0
    else:
        offset, sectors = result

    assert offset > 0
    assert sectors > 0

def test_chunk_location_empty_data() -> None:
    # Create region with valid header but no chunk data
    minimal_data = bytearray(8192) # 8KB of zeros (valid but empty header)
    region = Region(bytes(minimal_data))

    result = region.chunk_location(0, 0)
    assert result == (0, 0)

def test_chunk_location_raises_error() -> None:
    with pytest.raises(EmptyRegionFile):
        region = Region(b'')
        region.chunk_location(0, 0)

def test_chunk_data_existing_chunk() -> None:
    # set up variables
    empty_region = EmptyRegion(0, 0)
    chunk = EmptyChunk(0, 0)
    empty_region.add_chunk(chunk)

    # Convert to region
    region_bytes = empty_region.save()
    region = Region(region_bytes)

    # Call the function
    result = region.chunk_data(0, 0)
    assert result != None

def test_chunk_data_returns_none() -> None:
    pass

def test_chunk_data_empty_data() -> None:
    pass

def test_chunk_data_handle_unsupported_gzip() -> None:
    pass

def test_chunk_data_handle_empty_region() -> None:
    pass

def test_chunk_data_handle_unicode_decode_error() -> None:
    pass

def test_chunk_data_handle_corrupted_data() -> None:
    pass
